#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2010-2011 Carnegie Mellon University
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""A library for reading whole-slide images.

This package provides Python bindings for the OpenSlide library.  For
documentation on the OpenSlide API, see:

http://openslide.org/api/openslide_8h.html
"""

from __future__ import division
from collections import Mapping
from PIL import Image

from openslide import lowlevel

# For the benefit of library users
from openslide.lowlevel import OpenSlideError
from openslide._version import __version__

PROPERTY_NAME_COMMENT          = 'openslide.comment'
PROPERTY_NAME_VENDOR           = 'openslide.vendor'
PROPERTY_NAME_QUICKHASH1       = 'openslide.quickhash-1'
PROPERTY_NAME_BACKGROUND_COLOR = 'openslide.background-color'

class AbstractSlide(object):
    """The base class of a slide object."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @classmethod
    def can_open(cls, filename):
        """Return True if the specified file can be read."""
        raise NotImplementedError

    def close(self):
        """Close the slide."""
        raise NotImplementedError

    @property
    def layer_count(self):
        """The number of layers in the image."""
        raise NotImplementedError

    @property
    def layer_dimensions(self):
        """A list of (width, height) tuples, one for each layer of the image.

        layer_dimensions[n] contains the dimensions of layer n."""
        raise NotImplementedError

    @property
    def dimensions(self):
        """A (width, height) tuple for layer 0 of the image."""
        return self.layer_dimensions[0]

    @property
    def layer_downsamples(self):
        """A list of downsampling factors for each layer of the image.

        layer_downsample[n] contains the downsample factor of layer n."""
        raise NotImplementedError

    @property
    def properties(self):
        """Metadata about the image.

        This is a map: property name -> property value."""
        raise NotImplementedError

    @property
    def associated_images(self):
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image."""
        raise NotImplementedError

    def get_best_layer_for_downsample(self, downsample):
        """Return the best layer for displaying the given downsample."""
        raise NotImplementedError

    def read_region(self, location, layer, size):
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the layer 0
                  reference frame.
        layer:    the layer number.
        size:     (width, height) tuple giving the region size."""
        raise NotImplementedError

    def get_thumbnail(self, size):
        """Return a PIL.Image containing an RGB thumbnail of the image.

        size:     the maximum size of the thumbnail."""
        downsample = max(*[dim / thumb for dim, thumb in
                zip(self.dimensions, size)])
        layer = self.get_best_layer_for_downsample(downsample)
        tile = self.read_region((0, 0), layer, self.layer_dimensions[layer])
        # Apply on solid background
        bg_color = '#' + self.properties.get(PROPERTY_NAME_BACKGROUND_COLOR,
                'ffffff')
        thumb = Image.new('RGB', tile.size, bg_color)
        thumb.paste(tile, None, tile)
        thumb.thumbnail(size, Image.ANTIALIAS)
        return thumb


class OpenSlide(AbstractSlide):
    """An open whole-slide image.

    close() is called automatically when the object is deleted.  In
    addition, an OpenSlide object may be used as a context manager, and
    will be closed when exiting the context.

    If an operation fails, OpenSlideError is raised.  Note that OpenSlide
    has latching error semantics: once OpenSlideError is raised, all future
    operations on the OpenSlide object, other than close(), will fail.
    """

    def __init__(self, filename):
        """Open a whole-slide image."""
        AbstractSlide.__init__(self)
        self._osr = lowlevel.open(filename)

    @classmethod
    def can_open(cls, filename):
        """Return True if OpenSlide can read the specified file."""
        return lowlevel.can_open(filename)

    def close(self):
        """Close the OpenSlide object."""
        lowlevel.close(self._osr)

    @property
    def layer_count(self):
        """The number of layers in the image."""
        return lowlevel.get_layer_count(self._osr)

    @property
    def layer_dimensions(self):
        """A list of (width, height) tuples, one for each layer of the image.

        layer_dimensions[n] contains the dimensions of layer n."""
        return tuple(lowlevel.get_layer_dimensions(self._osr, i)
                for i in range(self.layer_count))

    @property
    def layer_downsamples(self):
        """A list of downsampling factors for each layer of the image.

        layer_downsample[n] contains the downsample factor of layer n."""
        return tuple(lowlevel.get_layer_downsample(self._osr, i)
                for i in range(self.layer_count))

    @property
    def properties(self):
        """Metadata about the image.

        This is a map: property name -> property value."""
        return _PropertyMap(self._osr)

    @property
    def associated_images(self):
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image.

        Unlike in the C interface, the images accessible via this property
        are not premultiplied."""
        return _AssociatedImageMap(self._osr)

    def get_best_layer_for_downsample(self, downsample):
        """Return the best layer for displaying the given downsample."""
        return lowlevel.get_best_layer_for_downsample(self._osr, downsample)

    def read_region(self, location, layer, size):
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the layer 0
                  reference frame.
        layer:    the layer number.
        size:     (width, height) tuple giving the region size.

        Unlike in the C interface, the image data returned by this
        function is not premultiplied."""
        return lowlevel.read_region(self._osr, location[0], location[1],
                layer, size[0], size[1])


class _OpenSlideMap(Mapping):
    def __init__(self, osr):
        self._osr = osr

    def __repr__(self):
        return repr(dict(self))

    def __len__(self):
        return len(self.keys())

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        raise NotImplementedError()


class _PropertyMap(_OpenSlideMap):
    def keys(self):
        return lowlevel.get_property_names(self._osr)

    def __getitem__(self, key):
        v = lowlevel.get_property_value(self._osr, key)
        if v is None:
            raise KeyError()
        return v


class _AssociatedImageMap(_OpenSlideMap):
    def keys(self):
        return lowlevel.get_associated_image_names(self._osr)

    def __getitem__(self, key):
        if key not in self.keys():
            raise KeyError()
        return lowlevel.read_associated_image(self._osr, key)


class ImageSlide(AbstractSlide):
    """A wrapper for a PIL.Image that provides the OpenSlide interface."""

    def __init__(self, file):
        """Open an image file.

        file can be a filename or a PIL.Image."""
        AbstractSlide.__init__(self)
        if isinstance(file, Image.Image):
            self._image = file
        else:
            self._image = Image.open(file)

    @classmethod
    def can_open(cls, filename):
        """Return True if PIL can read the specified file."""
        try:
            Image.open(filename)
            return True
        except IOError:
            return False

    def close(self):
        """Close the slide object."""
        self._image = None

    @property
    def layer_count(self):
        """The number of layers in the image."""
        return 1

    @property
    def layer_dimensions(self):
        """A list of (width, height) tuples, one for each layer of the image.

        layer_dimensions[n] contains the dimensions of layer n."""
        return (self._image.size,)

    @property
    def layer_downsamples(self):
        """A list of downsampling factors for each layer of the image.

        layer_downsample[n] contains the downsample factor of layer n."""
        return (1.0,)

    @property
    def properties(self):
        """Metadata about the image.

        This is a map: property name -> property value."""
        return {}

    @property
    def associated_images(self):
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image."""
        return {}

    def get_best_layer_for_downsample(self, _downsample):
        """Return the best layer for displaying the given downsample."""
        return 0

    def read_region(self, location, layer, size):
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the layer 0
                  reference frame.
        layer:    the layer number.
        size:     (width, height) tuple giving the region size."""
        if layer != 0:
            raise OpenSlideError("Invalid layer")
        if ['fail' for s in size if s < 0]:
            raise OpenSlideError("Size %s must be non-negative" % (size,))
        # Any corner of the requested region may be outside the bounds of
        # the image.  Create a transparent tile of the correct size and
        # paste the valid part of the region into the correct location.
        image_topleft = [max(0, min(l, limit - 1))
                    for l, limit in zip(location, self._image.size)]
        image_bottomright = [max(0, min(l + s - 1, limit - 1))
                    for l, s, limit in zip(location, size, self._image.size)]
        tile = Image.new("RGBA", size, (0,) * 4)
        if not ['fail' for tl, br in zip(image_topleft, image_bottomright)
                if br - tl < 0]:  # "< 0" not a typo
            # Crop size is greater than zero in both dimensions.
            # PIL thinks the bottom right is the first *excluded* pixel
            crop = self._image.crop(image_topleft +
                    [d + 1 for d in image_bottomright])
            tile_offset = tuple(il - l for il, l in
                    zip(image_topleft, location))
            tile.paste(crop, tile_offset)
        return tile


def open_slide(filename):
    """Open a whole-slide or regular image.

    Return an OpenSlide object for whole-slide images and an ImageSlide
    object for other types of images."""
    if OpenSlide.can_open(filename):
        return OpenSlide(filename)
    else:
        return ImageSlide(filename)


if __name__ == '__main__':
    import sys
    print "OpenSlide can open:", OpenSlide.can_open(sys.argv[1])
    print "PIL can open:", ImageSlide.can_open(sys.argv[1])
    with open_slide(sys.argv[1]) as _slide:
        print "Dimensions:", _slide.dimensions
        print "Layers:", _slide.layer_count
        print "Layer dimensions:", _slide.layer_dimensions
        print "Layer downsamples:", _slide.layer_downsamples
        print "Properties:", _slide.properties
        print "Associated images:", _slide.associated_images
